# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for random scalars."""

import unittest
from pyglove.ext.scalars import randoms as scalars


class RandomScalarsTest(unittest.TestCase):
  """Random scalars tests."""

  def test_uniform(self):
    sv = scalars.Uniform(seed=1)
    self.assertEqual(sv(0), 0.13436424411240122)
    self.assertEqual(sv(0), 0.8474337369372327)

    sv = scalars.Uniform(1, 10, seed=1)
    self.assertEqual(sv(0), 3)
    self.assertEqual(sv(0), 10)

    with self.assertRaisesRegex(
        ValueError,
        '`low` must be less or equal than `high`.'):
      scalars.Uniform(10, 1, seed=1)

  def test_triangular(self):
    sv = scalars.Triangular(0.0, 1.0, 0.9, seed=1)
    self.assertEqual(sv(0), 0.34774677525630787)
    self.assertEqual(sv(0), 0.8733214547023962)

    sv = scalars.Triangular(10, 20, seed=1)
    self.assertEqual(sv(0), 12)
    self.assertEqual(sv(0), 17)

  def test_gaussian(self):
    sv = scalars.Gaussian(1.0, 0.2, seed=1)
    self.assertEqual(sv(0), 1.2576369506310927)
    self.assertEqual(sv(0), 1.2898891217399542)

  def test_normal(self):
    sv = scalars.Normal(1.0, 0.2, seed=1)
    self.assertEqual(sv(0), 1.1214911715287412)
    self.assertEqual(sv(0), 0.997154910897843)

  def test_log_normal(self):
    sv = scalars.LogNormal(1.0, 0.2, seed=1)
    self.assertEqual(sv(0), 3.0694278358084994)
    self.assertEqual(sv(0), 2.710559065635824)


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic Hill-Climbing Algorithm."""

from typing import Optional

import pyglove.core as pg
from pyglove.ext.evolution import base
from pyglove.ext.evolution import mutators
from pyglove.ext.evolution import selectors


def hill_climb(mutator=mutators.Uniform(),
               batch_size: int = 1,
               init_population_size: int = 1,
               seed: Optional[int] = None) -> base.Evolution:
  """Hill-Climbing algorithm, with an extra batched setting.

  Batched setting was shown to be effective in
  https://arxiv.org/pdf/1911.06317.pdf and https://arxiv.org/pdf/2003.01239.pdf,
  especially in noisy objective settings.

  Args:
    mutator: Mutator to use.
    batch_size: Number of mutations of the current best.
    init_population_size: Initial population size (randomly generated).
    seed: Random seed.

  Returns:
    An `Evolution` object.
  """
  return base.Evolution(
      selectors.Top(1) >> (mutator * batch_size),   # pytype: disable=unsupported-operands
      population_init=(pg.geno.Random(seed), init_population_size),
      population_update=selectors.Top(1))




# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pluggable file system."""

import abc
import io
import os
from typing import Any, Literal, Optional, Union


class File(metaclass=abc.ABCMeta):
  """Interface for a file."""

  @abc.abstractmethod
  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    """Reads the content of the file."""

  @abc.abstractmethod
  def readline(self) -> Union[str, bytes]:
    """Reads the next line."""

  @abc.abstractmethod
  def write(self, content: Union[str, bytes]) -> None:
    """Writes the content of the file."""

  @abc.abstractmethod
  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    """Changes the current position of the file.

    Args:
      offset: Offset from the position to a reference point.
      whence: The reference point, with 0 meaning the beginning of the file,
        1 meaning the current position, or 2 meaning the end of the file.

    Returns:
      The position from the beginning of the file.
    """

  @abc.abstractmethod
  def tell(self) -> int:
    """Returns the current position of the file."""

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the file."""

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, exc_value, traceback
    self.close()


class FileSystem(metaclass=abc.ABCMeta):
  """Interface for a file system."""

  @abc.abstractmethod
  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    """Opens a file with a path."""

  @abc.abstractmethod
  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Returns True if a path exists."""

  @abc.abstractmethod
  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    """Lists all files or sub-directories."""

  @abc.abstractmethod
  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Returns True if a path is a directory."""

  @abc.abstractmethod
  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    """Makes a directory based on a path."""

  @abc.abstractmethod
  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = False,
  ) -> None:
    """Makes a directory chain based on a path."""

  @abc.abstractmethod
  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    """Removes a file based on a path."""

  @abc.abstractmethod
  def rmdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Removes a directory based on a path."""

  @abc.abstractmethod
  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    """Removes a directory chain based on a path."""


def _resolve_path(path: Union[str, os.PathLike[str]]) -> str:
  if isinstance(path, str):
    return path
  elif hasattr(path, '__fspath__'):
    return path.__fspath__()
  else:
    raise ValueError(f'Unsupported path: {path!r}.')


#
# The standard file system.
#


class StdFile(File):
  """The standard file."""

  def __init__(self, file_object) -> None:
    super().__init__()
    self._file_object = file_object

  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    return self._file_object.read(size)

  def readline(self) -> Union[str, bytes]:
    return self._file_object.readline()

  def write(self, content: Union[str, bytes]) -> None:
    self._file_object.write(content)

  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    return self._file_object.seek(offset, whence)

  def tell(self) -> int:
    return self._file_object.tell()

  def close(self) -> None:
    self._file_object.close()


class StdFileSystem(FileSystem):
  """The standard file system."""

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    return StdFile(io.open(path, mode, **kwargs))

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    return os.path.exists(path)

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    return os.listdir(path)

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    return os.path.isdir(path)

  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    os.mkdir(path, mode)

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = False,
  ) -> None:
    os.makedirs(path, mode, exist_ok)

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    os.remove(path)

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:
    os.rmdir(path)

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    os.removedirs(path)


#
# Memory file system.
#


class MemoryFile(File):
  """Memory file."""

  def __init__(self, buffer: io.IOBase):
    super().__init__()
    self._buffer = buffer
    self._pos = 0

  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    return self._buffer.read(size)

  def readline(self) -> Union[str, bytes]:
    return self._buffer.readline()

  def write(self, content: Union[str, bytes]) -> None:
    self._buffer.write(content)

  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    return self._buffer.seek(offset, whence)

  def tell(self) -> int:
    return self._buffer.tell()

  def close(self) -> None:
    self.seek(0)


class MemoryFileSystem(FileSystem):
  """The in-memory file system."""

  def __init__(self, prefix: str = '/mem/'):
    super().__init__()
    self._root = {}
    self._prefix = prefix

  def _internal_path(self, path: Union[str, os.PathLike[str]]) -> str:
    return '/' + _resolve_path(path).lstrip(self._prefix)

  def _locate(self, path: Union[str, os.PathLike[str]]) -> Any:
    current = self._root
    for x in self._internal_path(path).split('/'):
      if not x:
        continue
      if x not in current:
        return None
      current = current[x]
    return current

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    file = self._locate(path)
    if isinstance(file, dict):
      raise IsADirectoryError(path)
    if 'w' in mode and file is None:
      parent_dir, name = self._parent_and_name(path)
      if isinstance(parent_dir, dict):
        buffer = io.BytesIO() if 'b' in mode else io.StringIO()
        file = MemoryFile(buffer)
        parent_dir[name] = file

    if file is None:
      raise FileNotFoundError(path)
    return file

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    return self._locate(path) is not None

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    d = self._locate(path)
    if not isinstance(d, dict):
      raise FileNotFoundError(path)
    return list(d.keys())

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    return isinstance(self._locate(path), dict)

  def _parent_and_name(
      self, path: Union[str, os.PathLike[str]]
  ) -> tuple[dict[str, Any], str]:
    path = _resolve_path(path)
    rpos = path.rfind('/')
    assert rpos >= 0, path
    name = path[rpos + 1:]
    parent_dir = self._locate(path[:rpos])
    if parent_dir is None:
      raise FileNotFoundError(path)
    return parent_dir, name

  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    del mode
    parent_dir, name = self._parent_and_name(path)
    if name in parent_dir:
      raise FileExistsError(path)
    parent_dir[name] = {}

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = False,
  ) -> None:
    del mode
    current = self._root
    dirs = self._internal_path(path).split('/')
    for i, x in enumerate(dirs):
      if not x:
        continue
      entry = current.get(x)
      if entry is None:
        entry = {}
        current[x] = entry
      elif isinstance(entry, dict) and i == len(dirs) - 1 and not exist_ok:
        raise FileExistsError(path)
      elif not isinstance(entry, dict):
        raise NotADirectoryError(path)
      current = entry

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    parent_dir, name = self._parent_and_name(path)
    entry = parent_dir.get(name)
    if entry is None:
      raise FileNotFoundError(path)
    elif isinstance(entry, dict):
      raise IsADirectoryError(path)
    del parent_dir[name]

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:
    parent_dir, name = self._parent_and_name(path)
    entry = parent_dir.get(name)
    if entry is None:
      raise FileNotFoundError(path)
    elif not isinstance(entry, dict):
      raise NotADirectoryError(path)
    elif entry:
      raise OSError(f'Directory not empty: {path!r}')
    del parent_dir[name]

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    def _rmdir(dir_dict, subpath: str) -> bool:
      if not subpath:
        if dir_dict:
          raise OSError(f'Directory not empty: {path!r}')
        return True
      subpath = subpath.lstrip('/')
      pos = subpath.find('/')
      if pos < 0:
        pos = len(subpath)
      name = subpath[:pos]
      subpath = subpath[pos + 1:] if pos < len(subpath) else ''
      if name not in dir_dict:
        raise FileNotFoundError(path)
      elif not isinstance(dir_dict[name], dict):
        raise NotADirectoryError(path)
      if _rmdir(dir_dict[name], subpath):
        del dir_dict[name]
      return not dir_dict
    _rmdir(self._root, self._internal_path(path))


class _FileSystemRegistry:
  """File system registry."""

  def __init__(self):
    self._filesystems = []

  def add(self, prefix, fs: FileSystem) -> None:
    """Gets the file system applicable for path."""
    self._filesystems.append((prefix, fs))
    self._filesystems.sort(key=lambda x: x[0], reverse=True)

  def get(self, path: Union[str, os.PathLike[str]]) -> FileSystem:
    """Gets the file system for a path."""
    path = _resolve_path(path)
    for prefix, fs in self._filesystems:
      if path.startswith(prefix):
        return fs
    return StdFileSystem()


_fs = _FileSystemRegistry()


def add_file_system(prefix: str, fs: FileSystem) -> None:
  """Adds a file system with a prefix."""
  _fs.add(prefix, fs)


# Register a memory file system.
add_file_system('/mem/', MemoryFileSystem('/mem/'))


#
# APIs for file IO.
#


def open(path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs) -> File:  # pylint:disable=redefined-builtin
  """Opens a file with a path."""
  return _fs.get(path).open(path, mode, **kwargs)


def readfile(
    path: Union[str, os.PathLike[str]],
    mode: str = 'r',
    nonexist_ok: bool = False,
    **kwargs,
) -> Union[bytes, str, None]:
  """Reads content from a file."""
  try:
    with _fs.get(path).open(path, mode=mode, **kwargs) as f:
      return f.read()
  except FileNotFoundError as e:
    if nonexist_ok:
      return None
    raise e


def writefile(
    path: Union[str, os.PathLike[str]],
    content: Union[str, bytes],
    *,
    mode: str = 'w',
    **kwargs,
) -> None:
  """Writes content to a file."""
  with _fs.get(path).open(path, mode=mode, **kwargs) as f:
    f.write(content)


def rm(path: Union[str, os.PathLike[str]]) -> None:
  """Removes a file."""
  _fs.get(path).rm(path)


def path_exists(path: Union[str, os.PathLike[str]]) -> bool:
  """Returns True if path exists."""
  return _fs.get(path).exists(path)


def listdir(
    path: Union[str, os.PathLike[str]], fullpath: bool = False
) -> list[str]:  # pylint: disable=redefined-builtin
  """Lists all files or sub-directories under a dir."""
  entries = _fs.get(path).listdir(path)
  if fullpath:
    return [os.path.join(path, entry) for entry in entries]
  return entries


def isdir(path: Union[str, os.PathLike[str]]) -> bool:
  """Returns True if path is a directory."""
  return _fs.get(path).isdir(path)


def mkdir(path: Union[str, os.PathLike[str]], mode: int = 0o777) -> None:
  """Makes a directory."""
  _fs.get(path).mkdir(path, mode=mode)


def mkdirs(
    path: Union[str, os.PathLike[str]],
    mode: int = 0o777,
    exist_ok: bool = False,
) -> None:
  """Makes a directory chain."""
  _fs.get(path).mkdirs(path, mode=mode, exist_ok=exist_ok)


def rmdir(path: Union[str, os.PathLike[str]]) -> bool:
  """Removes a directory."""
  return _fs.get(path).rmdir(path)


def rmdirs(path: Union[str, os.PathLike[str]]) -> bool:
  """Removes a directory chain until a parent directory is not empty."""
  return _fs.get(path).rmdirs(path)



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for evolutionary algorithms."""

import random
import time
import unittest

import pyglove.core as pg
from pyglove.ext.evolution import base
from pyglove.ext.evolution import nsga2_lib as nsga2


def get_trivial_search_space():
  """Trivial search space.

  Each point in the space is a value in [0, 1].

  Returns:
    A tunable value.
  """
  return pg.floatv(0.0, 1.0)


def trivial_reward(example):
  """Reward for the trivial search space.

  The reward (i.e. fitness) is the value itself. The goal of the search,
  therefore, is to find the value 1.

  Args:
    example: a materialized value.

  Returns:
    The corresponding reward.
  """
  return example


def simple_two_objective_reward(example):
  """Reward for the trivial search space.

  The reward (i.e. fitness) is a 2-element list. The goal of the search,
  therefore, is to find the pareto frontier in simple_two_objective_pareto
  function.

  Args:
    example: a materialized value.

  Returns:
    A 2-element list.
  """
  num = int(example * 10) % 9 + 1  # Maps to [1 ... 9].

  return [num, 10 - num]


def simple_two_objective_pareto():
  """The pareto frontier of simple_two_objective_reward function."""
  return [[9, 1], [8, 2], [7, 3], [6, 4], [5, 5], [4, 6], [3, 7], [2, 8],
          [1, 9]]


def multi_frontier_two_objective_reward(example):
  """Reward for the trivial search space.

  The reward (i.e. fitness) is a 2-element list. The goal of the search,
  therefore, is to find the pareto frontier in
  multi_frontier_two_objective_pareto function.

  Args:
    example: a materialized value.

  Returns:
    A 2-element list.
  """
  int_val = int(example * 10)
  if int_val >= 0 and int_val < 3:
    return [int_val, 10 - int_val]  # Maps to [0, 10], [1, 9], [2, 8]
  elif int_val >= 3 and int_val < 7:
    return [int_val * 10, 100 - int_val * 10
           ]  # Maps to [30, 70], [40, 60], [50, 50], [60, 40]
  else:
    return [int_val, 10 - int_val]  # Maps to [7, 3], [8, 2], [9, 1], [10, 0]


def multi_frontier_two_objective_pareto():
  """The pareto frontier of multi_frontier_two_objective_reward function."""
  return [[30, 70], [40, 60], [50, 50], [60, 40]]


def get_trivial_hash(search_space, algo):
  hashed_value = 0
  for example, feedback in pg.iter(search_space, 30, algo):
    hashed_value ^= int(example * 1000000)
    feedback([example])
  return hashed_value


@pg.members([
    ('seed', pg.typing.Int().noneable()),
])
class TrivialMutator(base.Mutator):
  """Mutator for trivial search space.

  Mutations can only change the value by a small amount.
  """

  def _on_bound(self):
    super()._on_bound()
    self._random = random if self.seed is None else random.Random(self.seed)

  def mutate(self, dna, step):
    del step
    value = dna.value + self._random.uniform(-0.01, 0.01)
    if value < 0.0:
      value = 0.0
    if value > 1.0:
      value = 1.0
    return pg.DNA(value)


class Nsga2Test(unittest.TestCase):

  def _contains(self, outgoing_population, pareto_frontier):
    """Returns True if outgoing_population contains pareto_frontier."""
    # Record the number of occurrence of each element in
    # expected_pareto_frontier.
    cache = [False] * len(pareto_frontier)
    for individual in outgoing_population:
      if pareto_frontier.count(base.get_fitness(individual)):
        cache[pareto_frontier.index(base.get_fitness(individual))] = True
      if False not in cache:
        return True
    return False

  def test_dominates(self):
    self.assertTrue(nsga2.dominates(tuple((1.0, 2.0)), tuple((1.0, 1.0))))
    self.assertFalse(nsga2.dominates(tuple((1.0, 2.0)), tuple((2.0, 1.0))))
    with self.assertRaises(ValueError):
      nsga2.dominates(tuple((1.0, 2.0, 3.0)), tuple((1.0, 2.0)))

  def test_fast_nondominated_sort(self):
    ind1 = base.set_fitness(pg.DNA(None), tuple((1.0, 1.0)))
    ind2 = base.set_fitness(pg.DNA(None), tuple((1.0, 2.0)))
    ind3 = base.set_fitness(pg.DNA(None), tuple((3.0, 4.0)))
    ind4 = base.set_fitness(pg.DNA(None), tuple((4.0, 3.0)))

    self.assertEqual(
        nsga2.nondominated_sort()([ind1, ind2]), [[ind2], [ind1]])
    self.assertEqual(
        nsga2.nondominated_sort()([ind3, ind4]), [[ind3, ind4]])
    self.assertEqual(
        nsga2.nondominated_sort()([ind1, ind2, ind3, ind4]),
        [[ind3, ind4], [ind2], [ind1]])

  def test_crowding_distance_sort(self):
    ind1 = base.set_fitness(pg.DNA(None), tuple((100.0, 1.0)))
    ind2 = base.set_fitness(pg.DNA(None), tuple((50.0, 10.0)))
    ind3 = base.set_fitness(pg.DNA(None), tuple((30.0, 30.0)))
    ind4 = base.set_fitness(pg.DNA(None), tuple((10.0, 50.0)))
    ind5 = base.set_fitness(pg.DNA(None), tuple((1.0, 100.0)))

    self.assertEqual(
        nsga2.crowding_distance_sort()([ind1, ind2, ind3, ind4, ind5]),
        [ind1, ind5, ind2, ind4, ind3])

  def test_integration_single_objective(self):
    # Set up search space.
    search_space = get_trivial_search_space()

    # Search algorithm.
    algo = nsga2.nsga2(TrivialMutator(), 20)
    self.assertTrue(algo.multi_objective)

    # Search.
    best_reward = None
    iters = 0
    start_time = time.time()
    while True:
      for example, feedback in pg.iter(search_space, 100, algo):
        reward = trivial_reward(example)
        feedback([reward])
        if best_reward is None or reward > best_reward:
          best_reward = reward
        iters += 1
        if reward >= 1.0:
          break
      if reward >= 1.0:
        break
      if time.time() - start_time > 300.0:
        self.fail('Took too long to find a solution.')

  def test_integration_simple_two_objectives(self):
    # Set up search space.
    search_space = get_trivial_search_space()

    # Set up expected pareto frontier.
    expected_pareto_frontier = simple_two_objective_pareto()

    # Search algorithm.
    algo = nsga2.nsga2(TrivialMutator(), 20)

    # Search.
    iters = 0
    start_time = time.time()
    while True:
      for example, feedback in pg.iter(search_space, 100, algo):
        reward = simple_two_objective_reward(example)
        feedback(reward)
        iters += 1
        if self._contains(
            algo.global_state.get('elites', []), expected_pareto_frontier):
          break
      if self._contains(
          algo.global_state.get('elites', []), expected_pareto_frontier):
        break
      if time.time() - start_time > 300.0:
        self.fail('Took too long to find a solution.')

  def test_integration_with_two_layered_objectives(self):
    # Set up search space.
    search_space = get_trivial_search_space()

    # Set up expected pareto frontier.
    expected_pareto_frontier = multi_frontier_two_objective_pareto()

    # Search algorithm.
    algo = nsga2.nsga2(TrivialMutator(), 20)

    # Search.
    iters = 0
    start_time = time.time()
    while True:
      for example, feedback in pg.iter(search_space, 100, algo):
        reward = multi_frontier_two_objective_reward(example)
        feedback(reward)
        iters += 1
        if self._contains(
            algo.global_state.get('elites', []), expected_pareto_frontier):
          break
      if self._contains(
          algo.global_state.get('elites', []), expected_pareto_frontier):
        break
      if time.time() - start_time > 300.0:
        self.fail('Took too long to find a solution.')

  def test_permanence(self):
    search_space = get_trivial_search_space()
    algo = nsga2.nsga2(TrivialMutator(seed=1), 10, seed=1)

    # If a CL causes the following assert to fail, it means that the CL is
    # causing a difference in the behavior of the evolutionary algorithms. If
    # this is expected (e.g. a change in the random number generator), then
    # simply update the hash to the new value.
    self.assertEqual(get_trivial_hash(search_space, algo), 264567)


if __name__ == '__main__':
  unittest.main()
